module class_Pila

implicit none

type op_val
   character(len=1) :: oper
   real(kind=8)     :: val
end type op_val

type Pila
   integer                   :: lung
   integer                   :: lmax
   type(op_val), allocatable :: elem(:)
end type Pila

contains

   !--------------------------------------------------------------------------------------

   function Crea_Pila(lmax_in,ierr) result(P)

   ! Ritorna ierr .ne. 0 in caso di errore di allocazione

   implicit none

   integer, intent(in)  :: lmax_in
   integer, intent(out) :: ierr
   type(Pila)           :: P

   P%lung = 0
   P%lmax = lmax_in
   allocate(P%elem(lmax_in),stat=ierr)

   end function Crea_Pila

   !--------------------------------------------------------------------------------------

   function Push(elem_in,P_IO) result(push_ok)

   implicit none

   type(op_val), intent(in)  :: elem_in
   type(Pila), intent(inout) :: P_IO
   logical                   :: push_ok

   if (P_IO%lung .eq. P_IO%lmax) then
      ! La pila e' piena, esci con condizione di errore
      push_ok = .False.
      return
   endif

   P_IO%lung = P_IO%lung + 1
   P_IO%elem(P_IO%lung)%oper = elem_in%oper
   P_IO%elem(P_IO%lung)%val = elem_in%val
   push_ok = .True.

   end function Push

   !--------------------------------------------------------------------------------------

   function Pop(elem_out,P_IO) result(pop_ok)

   implicit none

   type(op_val), intent(out) :: elem_out
   type(Pila), intent(inout) :: P_IO
   logical                   :: pop_ok

   if (P_IO%lung .eq. 0) then
      ! La pila e' vuota, esci con condizione di errore
      pop_ok = .False.
      return
   endif

   elem_out%oper = P_IO%elem(P_IO%lung)%oper
   elem_out%val = P_IO%elem(P_IO%lung)%val
   P_IO%lung = P_IO%lung - 1
   pop_ok = .True.

   end function Pop

   !--------------------------------------------------------------------------------------

   recursive function Valuta_Pila(P_IO,valuta_ok) result(ris)

   implicit none

   type(Pila), intent(inout) :: P_IO
   logical, intent(out)      :: valuta_ok
   real(kind=8)              :: ris

   type(op_val)              :: elem
   real(kind=8)              :: A,B

   ! Prendo un elemento in cima alla pila
   valuta_ok = Pop(elem,P_IO)
   if (.not. valuta_ok) then
      ris = 0.0
      return
   endif

   ! Controllo il contenuto
   if (elem%oper .eq. 'v') then
      ! Se contiene il valore lo ritorno come output
      ris = elem%val
   else
      ! Se contiene un'operazione valuto i due operandi
      ris = 0.0
      B = Valuta_Pila(P_IO,valuta_ok)
      if (.not. valuta_ok) return
      A = Valuta_Pila(P_IO,valuta_ok)
      if (.not. valuta_ok) return

      if (elem%oper .eq. '+') then
         ris = A + B
      elseif (elem%oper .eq. '-') then
         ris = A - B
      elseif (elem%oper .eq. 'x') then
         ris = A * B
      elseif (elem%oper .eq. 'd') then
         ris = A / B
      endif
   endif

   end function Valuta_Pila

   !--------------------------------------------------------------------------------------

   subroutine stampa_Pila(P_I)

   implicit none

   type(Pila), intent(in) :: P_I
   integer                :: i

   do i = 1,P_I%lung
      select case(P_I%elem(i)%oper)
         case('v')
            write(*,*) 'Value:',P_I%elem(i)%val
         case('+')
            write(*,*) '+  :  Addizione'
         case('-')
            write(*,*) '-  :  Sottrazione'
         case('x')
            write(*,*) 'x  :  Moltiplicazione'
         case('d')
            write(*,*) 'd  :  Divisione'
         case default
            write(*,*) P_I%elem(i)%oper,' caso non contemplato'
      end select
   enddo

   end subroutine Stampa_Pila

   !--------------------------------------------------------------------------------------

   function Distruggi_Pila(P_IO) result(ierr)

   implicit none

   type(Pila), intent(inout) :: P_IO
   integer                   :: ierr

   P_IO%lmax = 0
   P_IO%lung = 0
   deallocate(P_IO%elem,stat=ierr)

   end function Distruggi_Pila

   !--------------------------------------------------------------------------------------

end module class_Pila
